﻿#pragma once
// #include "Engine.h"


// #include "Jolt/Physics/Body/BodyCreationSettings.h"
// #include <Jolt/Physics/Collision/Shape/SphereShape.h>
// #include <Jolt/Physics/Body/BodyCreationSettings.h>
// #include "Jolt/Physics/Collision/Shape/BoxShape.h"
// #include "Jolt/Physics/Collision/Shape/SphereShape.h"
// #include "Jolt/Physics/Collision/Shape/CapsuleShape.h"


// using namespace JPH;
// Layer that objects can be in, determines which other objects it can collide with
// Typically you at least want to have 1 layer for moving bodies and 1 layer for static bodies, but you can have more
// layers if you want. E.g. you could have a layer for high detail collision (which is not used by the physics simulation
// but only if you do collision testing).
namespace Layers {
	static constexpr JPH::ObjectLayer NON_MOVING = 0;
	static constexpr JPH::ObjectLayer MOVING = 1;
	static constexpr JPH::ObjectLayer NUM_LAYERS = 2;
};

/// Class that determines if two object layers can collide
class ObjectLayerPairFilterImpl : public JPH::ObjectLayerPairFilter {
	public:
		virtual bool ShouldCollide(JPH::ObjectLayer inObject1, JPH::ObjectLayer inObject2) const override;
};

// Each broadphase layer results in a separate bounding volume tree in the broad phase. You at least want to have
// a layer for non-moving and moving objects to avoid having to update a tree full of static objects every frame.
// You can have a 1-on-1 mapping between object layers and broadphase layers (like in this case) but if you have
// many object layers you'll be creating many broad phase trees, which is not efficient. If you want to fine tune
// your broadphase layers define JPH_TRACK_BROADPHASE_STATS and look at the stats reported on the TTY.
namespace BroadPhaseLayers {
	static constexpr JPH::BroadPhaseLayer NON_MOVING(0);
	static constexpr JPH::BroadPhaseLayer MOVING(1);
	static constexpr JPH::uint NUM_LAYERS(2);
};


// BroadPhaseLayerInterface implementation
// This defines a mapping between object and broadphase layers.
class BPLayerInterfaceImpl final : public JPH::BroadPhaseLayerInterface {
	public:
		BPLayerInterfaceImpl();

		virtual JPH::uint GetNumBroadPhaseLayers() const override;

		virtual JPH::BroadPhaseLayer GetBroadPhaseLayer(JPH::ObjectLayer inLayer) const override;

		#if defined(JPH_EXTERNAL_PROFILE) || defined(JPH_PROFILE_ENABLED)
		virtual const char* GetBroadPhaseLayerName(JPH::BroadPhaseLayer inLayer) const override;
		#endif // JPH_EXTERNAL_PROFILE || JPH_PROFILE_ENABLED

	private:
		JPH::BroadPhaseLayer mObjectToBroadPhase[Layers::NUM_LAYERS];
};

/// Class that determines if an object layer can collide with a broadphase layer
class ObjectVsBroadPhaseLayerFilterImpl : public JPH::ObjectVsBroadPhaseLayerFilter {
	public:
		virtual bool ShouldCollide(JPH::ObjectLayer inLayer1, JPH::BroadPhaseLayer inLayer2) const override;
};

// using namespace std;

// An example contact listener
class MyContactListener : public JPH::ContactListener {
	public:
		// See: ContactListener
		virtual JPH::ValidateResult OnContactValidate(const JPH::Body& inBody1, const JPH::Body& inBody2,
			JPH::RVec3Arg inBaseOffset,
			const JPH::CollideShapeResult& inCollisionResult) override;

		virtual void OnContactAdded(const JPH::Body& inBody1, const JPH::Body& inBody2,
			const JPH::ContactManifold& inManifold,
			JPH::ContactSettings& ioSettings) override;

		virtual void OnContactPersisted(const JPH::Body& inBody1, const JPH::Body& inBody2,
			const JPH::ContactManifold& inManifold,
			JPH::ContactSettings& ioSettings) override;

		virtual void OnContactRemoved(const JPH::SubShapeIDPair& inSubShapePair) override;
};


struct RayCastResClosest {
	bool hit = false;
	float dist;
	glm::vec3 pos;
	glm::vec3 norm = glm::vec3(0, 0, 0);
	JPH::BodyID body_id;
};

// An example activation listener
class MyBodyActivationListener : public JPH::BodyActivationListener {
	public:
		virtual void OnBodyActivated(const JPH::BodyID& inBodyID, JPH::uint64 inBodyUserData) override;

		virtual void OnBodyDeactivated(const JPH::BodyID& inBodyID, JPH::uint64 inBodyUserData) override;
};

class Physics {
	private:
		inline void set_up_ray_cast(
			JPH::RRayCast& r_ray_cast,
			JPH::RayCastSettings& ray_cast_settings,
			glm::vec3 origin,
			glm::vec3 dir
		) {
			// r_ray_cast.mDirection = glm::vec3(dir.GetX(), dir.GetY(), dir.GetZ());
			r_ray_cast.mDirection = JPH::Vec3(dir.x, dir.y, dir.z);
			r_ray_cast.mOrigin = JPH::Vec3(origin.x, origin.y, origin.z);

			ray_cast_settings.mBackFaceMode = JPH::EBackFaceMode::CollideWithBackFaces;
			ray_cast_settings.mTreatConvexAsSolid = true;
		}

		JPH::uint step = 0;

	public:
		Physics();
		void tick();
		JPH::TempAllocatorImpl* temp_allocator;
		JPH::JobSystemThreadPool* job_system;

		BPLayerInterfaceImpl broad_phase_layer_interface;
		ObjectVsBroadPhaseLayerFilterImpl object_vs_broadphase_layer_filter;
		ObjectLayerPairFilterImpl object_vs_object_layer_filter;
		JPH::PhysicsSystem* physics_system;
		MyBodyActivationListener* body_activation_listener;
		MyContactListener* contact_listener;
		JPH::BodyInterface* body_interface;

		RayCastResClosest shape_cast_closest(
			JPH::Ref<JPH::Shape> shape,
			glm::vec3 origin,
			JPH::BodyFilter& inBodyFilter,
			JPH::BroadPhaseLayerFilter inBroadPhaseLayerFilter = {},
			JPH::ObjectLayerFilter inObjectLayerFilter = {},
			JPH::ShapeFilter inShapeFilter = {}
		) {

			// RayCastResClosest res;

			// const JPH::Vec3 cast_motion = JPH::Vec3(0, -15, 0);
			const JPH::Vec3 cast_motion = JPH::Vec3(0, 0, 0);
			JPH::RShapeCast r_shape_cast = JPH::RShapeCast{
				shape,
				JPH::Vec3::sReplicate(1),
				// JPH::RMat44::sTranslation(glm::to_jph(origin)),
					JPH::RMat44::sIdentity(),
				cast_motion
			};

			JPH::ShapeCastSettings settings = {
				.mBackFaceModeTriangles = JPH::EBackFaceMode::CollideWithBackFaces,
			};

			JPH::ClosestHitCollisionCollector<JPH::CastShapeCollector> collector;

			physics_system->GetNarrowPhaseQuery().CastShape(
				r_shape_cast, settings,
				// JPH::Vec3(0,0,0),
				glm::to_jph(origin),
				collector,
				{}, {}, inBodyFilter
			);
			RayCastResClosest res;
			if(collector.HadHit()) {
				JPH::ShapeCastResult hit = collector.mHit;

				// r_shape_cast.
				JPH::RVec3 hit_pos = r_shape_cast.GetPointOnRay(collector.mHit.mFraction);
				glm::vec3 hit_p_1 = glm::from_jph(hit.mContactPointOn1);
				glm::vec3 hit_p_2 = glm::from_jph(hit.mContactPointOn1);



				res.hit = true;
				res.body_id = hit.mBodyID2;
				res.pos = glm::from_jph(hit_pos);
				// res.dist = glm::length(dir) * collector.mHit.mFraction;
				res.dist = collector.mHit.mFraction;


			}
			return res;
		}

		// // JPH::AllHitCollisionCollector<>
		// // JPH::ClosestHitCollisionCollector
		inline RayCastResClosest RayCastClosest(
			glm::vec3 origin,
			glm::vec3 dir,
			JPH::BodyFilter& inBodyFilter,
			JPH::BroadPhaseLayerFilter inBroadPhaseLayerFilter = {},
			JPH::ObjectLayerFilter inObjectLayerFilter = {},
			JPH::ShapeFilter inShapeFilter = {}
		) {
			JPH::RRayCast r_ray_cast;
			JPH::RayCastSettings ray_cast_settings;

			JPH::ClosestHitCollisionCollector<JPH::CastRayCollector> collector;

			set_up_ray_cast(r_ray_cast, ray_cast_settings, origin, dir);

			this->physics_system->GetNarrowPhaseQuery().CastRay(
				r_ray_cast, ray_cast_settings, collector,
				inBroadPhaseLayerFilter, inObjectLayerFilter, inBodyFilter, inShapeFilter
			);
			RayCastResClosest res;

			if (collector.HadHit()) {
				JPH::RayCastResult hit = collector.mHit;
				JPH::RVec3 hit_pos = r_ray_cast.GetPointOnRay(collector.mHit.mFraction);

				res.hit = true;
				res.body_id = hit.mBodyID;
				res.pos = glm::vec3(hit_pos.GetX(), hit_pos.GetY(), hit_pos.GetZ());
				res.dist = glm::length(dir) * collector.mHit.mFraction;

				JPH::BodyLockRead lock(this->physics_system->GetBodyLockInterface(), hit.mBodyID);
				if (lock.Succeeded()) {
					const JPH::Body& hit_body = lock.GetBody();
					JPH::Vec3 normal = hit_body.GetWorldSpaceSurfaceNormal(hit.mSubShapeID2, hit_pos);
					res.norm = glm::vec3(normal.GetX(), normal.GetY(), normal.GetZ());
				}
			}
			return res;
		}


};
